package transformers

import (
	"fmt"
	"os"
	"strings"

	"github.com/johnkerl/miller/v6/pkg/cli"
	"github.com/johnkerl/miller/v6/pkg/lib"
	"github.com/johnkerl/miller/v6/pkg/mlrval"
	"github.com/johnkerl/miller/v6/pkg/types"
)

const verbNameSparsify = "sparsify"

var sparsifyOptions = []OptionSpec{
	{Flag: "-s", Arg: "{filler string}", Type: "string", Desc: "What values to remove. Defaults to the empty string."},
	{Flag: "-f", Arg: "{a,b,c}", Type: "csv-list", Desc: "Specify field names to be operated on; any other fields won't be modified. The default is to modify all fields."},
}

var SparsifySetup = TransformerSetup{
	Verb:         verbNameSparsify,
	UsageFunc:    transformerSparsifyUsage,
	ParseCLIFunc: transformerSparsifyParseCLI,
	IgnoresInput: false,
	Options:      sparsifyOptions,
}

func transformerSparsifyUsage(
	o *os.File,
) {
	fmt.Fprintf(o, "Usage: %s %s [options]\n", "mlr", verbNameSparsify)
	fmt.Fprint(o,
		`Unsets fields for which the key is the empty string (or, optionally, another
specified value). Only makes sense with output format not being CSV or TSV.
`)

	WriteVerbOptions(o, sparsifyOptions)

	fmt.Fprint(o,
		`Example: if input is a=1,b=,c=3 then output is a=1,c=3.
`)
}

func transformerSparsifyParseCLI(
	pargi *int,
	argc int,
	args []string,
	_ *cli.TOptions,
	doConstruct bool, // false for first pass of CLI-parse, true for second pass
) (RecordTransformer, error) {

	// Skip the verb name from the current spot in the mlr command line
	argi := *pargi
	verb := args[argi]
	argi++

	fillerString := ""
	var specifiedFieldNames []string = nil

	var err error
	for argi < argc /* variable increment: 1 or 2 depending on flag */ {
		opt := args[argi]
		if !strings.HasPrefix(opt, "-") {
			break // No more flag options to process
		}
		if args[argi] == "--" {
			break // All transformers must do this so main-flags can follow verb-flags
		}
		argi++

		switch opt {
		case "-h", "--help":
			transformerSparsifyUsage(os.Stdout)
			return nil, cli.ErrHelpRequested

		case "-s":
			fillerString, err = cli.VerbGetStringArg(verb, opt, args, &argi, argc)
			if err != nil {
				return nil, err
			}

		case "-f":
			specifiedFieldNames, err = cli.VerbGetStringArrayArg(verb, opt, args, &argi, argc)
			if err != nil {
				return nil, err
			}

		default:
			return nil, cli.VerbErrorf(verb, "option \"%s\" not recognized", opt)
		}
	}

	*pargi = argi
	if !doConstruct { // All transformers must do this for main command-line parsing
		return nil, nil
	}

	transformer, err := NewTransformerSparsify(
		fillerString,
		specifiedFieldNames,
	)
	if err != nil {
		return nil, err
	}

	return transformer, nil
}

type TransformerSparsify struct {
	fillerString          string
	fieldNamesSet         map[string]bool
	recordTransformerFunc RecordTransformerFunc
}

func NewTransformerSparsify(
	fillerString string,
	specifiedFieldNames []string,
) (*TransformerSparsify, error) {

	tr := &TransformerSparsify{
		fillerString:  fillerString,
		fieldNamesSet: lib.StringListToSet(specifiedFieldNames),
	}
	if specifiedFieldNames == nil {
		tr.recordTransformerFunc = tr.transformAll
	} else {
		tr.recordTransformerFunc = tr.transformSome
	}

	return tr, nil
}

func (tr *TransformerSparsify) Transform(
	inrecAndContext *types.RecordAndContext,
	outputRecordsAndContexts *[]*types.RecordAndContext, // list of *types.RecordAndContext
	inputDownstreamDoneChannel <-chan bool,
	outputDownstreamDoneChannel chan<- bool,
) error {
	HandleDefaultDownstreamDone(inputDownstreamDoneChannel, outputDownstreamDoneChannel)

	if !inrecAndContext.EndOfStream {
		return tr.recordTransformerFunc(
			inrecAndContext,
			outputRecordsAndContexts,
			inputDownstreamDoneChannel,
			outputDownstreamDoneChannel,
		)
	} else {
		*outputRecordsAndContexts = append(*outputRecordsAndContexts, inrecAndContext) // end-of-stream marker
	}
	return nil
}

func (tr *TransformerSparsify) transformAll(
	inrecAndContext *types.RecordAndContext,
	outputRecordsAndContexts *[]*types.RecordAndContext, // list of *types.RecordAndContext
	inputDownstreamDoneChannel <-chan bool,
	outputDownstreamDoneChannel chan<- bool,
) error {
	inrec := inrecAndContext.Record
	outrec := mlrval.NewMlrmapAsRecord()

	for pe := inrec.Head; pe != nil; pe = pe.Next {
		if pe.Value.String() != tr.fillerString {
			// Reference OK because ownership transfer
			outrec.PutReference(pe.Key, pe.Value)
		}
	}

	outrecAndContext := types.NewRecordAndContext(outrec, &inrecAndContext.Context)
	*outputRecordsAndContexts = append(*outputRecordsAndContexts, outrecAndContext)
	return nil
}

func (tr *TransformerSparsify) transformSome(
	inrecAndContext *types.RecordAndContext,
	outputRecordsAndContexts *[]*types.RecordAndContext, // list of *types.RecordAndContext
	inputDownstreamDoneChannel <-chan bool,
	outputDownstreamDoneChannel chan<- bool,
) error {
	inrec := inrecAndContext.Record
	outrec := mlrval.NewMlrmapAsRecord()

	for pe := inrec.Head; pe != nil; pe = pe.Next {
		if tr.fieldNamesSet[pe.Key] {
			if pe.Value.String() != tr.fillerString {
				// Reference OK because ownership transfer
				outrec.PutReference(pe.Key, pe.Value)
			}
		} else {
			outrec.PutReference(pe.Key, pe.Value)
		}
	}

	outrecAndContext := types.NewRecordAndContext(outrec, &inrecAndContext.Context)
	*outputRecordsAndContexts = append(*outputRecordsAndContexts, outrecAndContext)
	return nil
}
